#!/usr/bin/perl -w
use strict;
use utf8;

#############################################
#
#	file:	MakeIndex_OSX.pl
#		(requires perl 5.8.1 [OSX 10.3])
#
#	project:	JTeXEdit
#
#	Copyright (c) 1999-2006 Norihito KOISO
#	All Rights Reserved
#
#############################################
#
#	使用法：
#		(1) AddYomiToIdx.pl で
#			漢字には次のようにかなで読みをつけておくか，
#				\index{アファインくうかん@アファイン空間}
#			または次のような一覧表を ***.idxd として作っておく．
#				よみ<tab>読み
#		(2) そして，latex が生成した "***.idx" を
#			MakeIndex_OSX.pl にドラッグすると "***.ind" を生成する．
#
#	標準では A,..,Z,あ,..,ん を区別して
#	\indexspace で区切るが，
#	TitleChar($) と MakeTitle($) を
#	定義し直せば A-Z を一括したりできる．
#
#############################################
#
#	履歴
#	2008-06-05: perl UTF-8 環境 に対応
#	2004-06-03:	perl 5.8.1 以上，ptex 専用, AddYomiToIdx.pl に非連動
#	2002-01-26:	AddYomiToIdx.pl に連動
#	2001-11-17:	Textures 用に ".=" を付加
#	2001-07-20:	hypertex に対応
#
#############################################

use encoding 'shift_jis';

# 単に \indexspace で区切る場合
sub TitleChar($) {
	my ($ch) = @_;
	return ($ch =~ tr/a-zA-Zぁ-ゖァ-ヶ/A-ZA-Zァ-ヶァ-ヶ/) ? $ch : '';
}
sub MakeTitle($) {
	my ($ch) = @_;
	return "\\indexspace";
}

# 例: A-Z をまとめ，各題を \itemtitle{ア} の形で出力
#sub TitleChar($) {
#	my ($ch) = @_;
#	if ($ch =~ tr/A-Z//) { # A-Z を一括
#		return 'A --- Z';
#	} elsif ($ch =~ tr/\t-~//) { # それ以外の 1 byte 文字を一括
#		return '記号';
#	} else { # カタカナ
#		return $ch;
#	}
#}
#sub MakeTitle($) {
#	my ($ch) = @_;
#	return "\\itemtitle \{$ch\}";
#}

# プログラム本体

my @Unknowns = ();

my %Dict = ();
sub ReadDict() {
	my @buff = <IN>;
	my @lines = ();
	while (@buff) {
		push(@lines, split(/\n|\r/, shift @buff));
	}
	while (@lines) {
		my $line = shift @lines;
		next if ($line =~ /^%/ or $line eq '');
		my @content = split(/\t/, $line);
		if (@content == 2) {
			$Dict{$content[1]} = $content[0];
		} else {
			warn "辞書の形式が正しくありません:\n$line\n";
		}
	}
}

sub Normalize($) {
	my ($str) = @_;
	$str =~ tr/a-zァ-ヶ/A-Zぁ-ゖ/;
	$str =~ tr
<ぁぃぅぇぉがぎぐげござじずぜぞだぢっづでどばぱびぴぶぷべぺぼぽゃゅょゎゐゑゔゕゖ>
<あいうえおかきくけこさしすせそたちつつてとははひひふふへへほほやゆよわいえふかけ>;
	$str =~ tr
/、。，．・：；？！゛゜´｀¨＾￣＿ヽヾゝゞ〃仝々〆〇ー―‐／＼〜‖｜…‥‘’“”（）〔〕［］｛｝〈〉《》「」『』【】＋−±×÷＝≠＜＞≦≧∞∴♂♀°′″℃￥＄¢£％＃＆＊＠§☆★○●◎◇◆□■△▲▽▼※〒→←↑↓〓∈∋⊆⊇⊂⊃∪∩∧∨¬⇒⇔∀∃∠⊥⌒∂∇≡≒≪≫√∽∝∵∫∬Å‰♯♭♪†‡¶◯─│┌┐┘└├┬┤┴┼━┃┏┓┛┗┣┳┫┻╋┠┯┨┷┿┝┰┥┸╂//d;
	# see DeleteIgnoreChars in AddYomiToIdx.pl
	return $str;
}

sub MakeYomi($) {
	my ($text) = @_;
	return '' unless (length($text));
	return $Dict{$text} if (defined($Dict{$text}));
	$text =~ tr/ァ-ヶ，．・／/ぁ-ゖ/d;
	$text =~ s/^([ぁ-ゖー]*)//;
	my $out = $1;
	return $out unless (length($text));
	while (length($text)) {
		$text =~ s/^([^ぁ-ゖー]+[ぁ-ゖー]*)//;
		my $word = $1;
		my $saveword = $word;
		my $postbuff = '';
		while (length($word)) {
			last if defined($Dict{$word});
			$word =~ s/(.)$//;
			$postbuff = $1 . $postbuff;
		}
		if ($word) {
			$out .= $Dict{$word};
			if ($postbuff =~ /[^ぁ-ゖー]/) {
				$text = $postbuff . $text;
			} else {
				$out .= $postbuff;
			}
		} else {
			push(@Unknowns, $saveword);
			$out .= $saveword;
		}
	}
	return $out;
}

sub AddYomiToTerms(@) {
	my @terms = @_;
	for my $term (@terms) {
		my $hadyomi = @$term > 1;
		my $str = ($hadyomi) ? shift @$term : $$term[0];
		$str =~ tr/ //s;
		($str) = $str =~ /^ ?(.*) ?$/;
		$str =~ tr/a-z/A-Z/;
		$str =~ s/([^\t-~]+)/Normalize(MakeYomi($1))/eg unless ($hadyomi);
		unshift(@$term, $str);
	}
	return @terms;
}

sub Protect($) {	# '"', '\\' を退避
	my ($str) = @_;
	$str =~ tr/\t/ /;
	$str =~ s/(["\\])(.)/"\t".ord($1).','.ord($2).';'/eg;
	return $str;
}
sub Unprotect($@) {
	my ($str, @buff) = @_;
	$str =~ s/\t(\d+),(\d+);/chr($1).chr($2)/eg;
	$str =~ s/"(.)/$1/eg;
	return $str;
}

sub ParseLineYomi($) {
	my ($body) = @_;
	my $protected = Protect($body);
	my @text = split(/\|/, $protected);
	return "| が複数あります:($body)" if (@text > 2);
	$body = shift(@text);
	my $cmd = @text ? $text[0] : '';
	return "| の後に ! があります:" if ($cmd =~ /!/);
	my @terms = map {[map {Unprotect($_)} split(/\@/, $_)]}
					split(/!/, $body);
	return "! が多すぎます:" if (@terms > 3);
	for (@terms) { return "\@ が多すぎます:" if (@$_ > 2); }
	return ($cmd, AddYomiToTerms(@terms));
}

my %YomiTab;
sub LineToInd($$@) {
	my ($cmd, $page, @terms) = @_;
	$page = "\\$cmd\{$page\}" if ($cmd);
	my $str = join(' ', map {$$_[0]} @terms);
	$YomiTab{$str} = [[map {$$_[1]} @terms]] unless (defined $YomiTab{$str});
	push @{$YomiTab{$str}}, $page;
}

sub MakeIdx() {
	print OUT "\\begin{theindex}\n\n";
	my ($preTitle, @preWord) = ('', '', '', '');
	for my $yomi (sort keys %YomiTab) {
		my ($words, @pages) = @{$YomiTab{$yomi}};
		$yomi =~ /(.)/;
		my $tt = TitleChar($1);
		print OUT "\n\t", MakeTitle($tt), "\n\n" if ($tt ne $preTitle);
		$preTitle = $tt;
		my @word = (@$words, '', '');
		my @out = ();
		push(@out, "\t\\item $word[0]") if ($word[0] ne $preWord[0]);
		push(@out, "\t\t\\subitem $word[1]")
			if (length($word[1]) and (@out or $word[1] ne $preWord[1]));
		push(@out, "\t\t\t\\subsubitem $word[2]") if (length($word[2]));
		print OUT join("\n", @out), ", ", join(", ", @pages), "\n";
		@preWord = @word;
	}
	print OUT "\n\\end{theindex}\n";
}

sub ParseBody($) {
	my ($body) = @_;
	my @text = split(/\|/, Protect($body));
	return "| が複数あります:($body)" if (@text > 2);
	$body = shift(@text);
	my $cmd = @text ? $text[0] : '';
	return "| の後に ! があります:" if ($cmd =~ /!/);
	my @terms = map {[map {Unprotect($_)} split(/\@/, $_)]}
					split(/!/, $body);
	return "! が多すぎます:" if (@terms > 3);
	for (@terms) {
		return "\@ が多すぎます:" if (@$_ > 2);
		return "\@ が先頭にあります:" unless $$_[0];
	}
	return ($cmd, AddYomiToTerms(@terms));
}

sub ParseFile() {
	while (defined(my $line = <IN>)) {
		chomp $line;
		if ($line =~ /^\\indexentry\{(.*)\}\{(\d+)\}$/) {
			my $page = $2;
			my ($cmd, @terms) = ParseBody($1);
			if (@terms) { LineToInd($cmd, $page, @terms); }
			else { warn "$cmd $line\n"; }
		}
	}
	MakeIdx();
}

print "\n**** MakeIndex_OSX.pl ****\n";

{
	die ".idx ファイルを指定して下さい．\n" unless @ARGV;
	my $idxFile = shift;
	my $dictFile = $idxFile.'d';
	print "辞書読込 < $dictFile\n";
	if (-e $dictFile) {
		open(IN, "<:encoding(shift_jis)", $dictFile) or die "辞書が開けません\n";
		ReadDict();
		close(IN);
	} else {
		print "辞書はありません\n";
	}
	my $outFile = $idxFile;
	$outFile =~ s/\.[^.:]*$/.ind/;
	print "処理開始: $idxFile\n > $outFile\n";
	open(IN, "<:encoding(shift_jis)", $idxFile) or die "ファイルが開けません:\n'$idxFile'\n";
	open(OUT, ">:encoding(shift_jis)", $outFile) or die "ファイルが開けません:\n'$outFile'\n";
	ParseFile();
	close(IN);
	close(OUT);
	print "終了．\n";
	if (@Unknowns) {
		my $msg = "以下の文字列が読めませんでした．\n";
		$msg .= "'$dictFile' に追加してください．\n";
		$msg .= join("\n", sort(@Unknowns)) . "\n";
		warn $msg;
	}
}
