#! /usr/bin/perl -w
use strict;
use utf8;
# this file must be utf-8 encoded

BEGIN {
my $dir = $0;
$dir =~ s|/[^/]*$||;
unshift(@INC, $dir);
}
use PListIO;

my $Root = shift;
if (! $Root) {
	$Root = $0;
	$Root =~ s|/[^/]*$|/..|;
}

sub ReadFile($) {
	my ($file) = @_;
	open(IN, "<:encoding(shift_jis)", "$Root/$file") or die "Can't open $Root/$file ";
	my $text = join("", <IN>);
	my @lines = split(/\n|\r/, $text);
	close(IN);
	return @lines;
}

sub Get_Files($$) {
	my ($file, $dir) = @_;
	my @files = ();
	push(@files, $file) if ($file and -e "$Root/$file");
	if (-e "$Root/$dir") {
		opendir(DIR, "$Root/$dir");
		my @items = sort(readdir(DIR));
		for (@items) {
			push(@files, "$dir/$_") unless /^\./;
		}
	}
	return @files;
}

sub AddToDict(\%$$) {
	my ($dict, $name, $item) = @_;
	$$dict{$name} = $item if ($item);
	TrashFile($name);
}

sub TrashFile($) {
	my ($name) = @_;
	my $trash = "$Root/-trash-";
	mkdir($trash, 0755) unless (-e $trash);
	rename("$Root/$name", "$trash/$name");
}

sub DecodeSlashRT($) {
	my ($str) = @_;
	$str =~ s|/r|\n|g;
	$str =~ s|/t|\t|g;
	return $str;
}

# CloseSymbol
#	^[^%](.+)\t(.+)$
#	decode '/r', '/t', '/|'
sub Parse_CloseSymbol() {
	return undef unless(-e "$Root/CloseSymbol");
	my @lines = ReadFile("CloseSymbol");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		$line =~ /^(.+)\t(.+)$/ or next;
		push(@list, {'key' => $1, 'close' => DecodeSlashRT($2)});
	}
	return {
		'help' => '/| を選択する．$ と $$ はここでは設定できない．',
		'list' => \@list
	};
}

# ContextEditMenu
#	^[^%].*$
sub Parse_ContextEditMenu() {
	return undef unless(-e "$Root/ContextEditMenu");
	my @lines = ReadFile("ContextEditMenu");
	my @list = ();
	for my $line (@lines) {
		push(@list, $line) if ($line !~ /^%/ and $line ne '');
	}
	return {
		'help' => '存在しない項目は無視されるので英語版と兼用',
		'list' => \@list
	};
}

# IndexMenu
#	^.*$
#	\t_property_ rgb=r,g,b; face=NBIUOSCE;
sub Parse_IndexMenu() {
	return undef unless(-e "$Root/IndexMenu");
	my @lines = ReadFile("IndexMenu");
	my @list = ();
	for my $line (@lines) {
		my %hash;
		if ($line =~ /^(.+)\s_property_\s+(.+)$/) {
			$line = $1;
			my $prop = $2;
			if ($prop =~ /rgb=([\d.]+),([\d.]+),([\d.]+);/) {
				my ($r,$g,$b) = ($1+0,$2+0,$3+0);
				$hash{'rgb'} = "$r,$g,$b";
			}
			if ($prop =~ /face=([NBIUOSCE]+);/) {
				$hash{'face'} = $1;
			}
		}
		$hash{'text'} = $line;
		push(@list, \%hash);
	}
	return {
		'help' => '色は rgb => r,g,b で，変形は face => NBIUOSCE で指定する．',
		'list' => \@list
	};
}

# MacroMenu
#	item = (\t*)(.+)\t(.+)\t(.+)	(indent, title, menukey, text)
#	hiermenu = (\t*)([^\t]+)		(indent, title)
#	separation = -					(title)
#	decode '/r', '/t', '/|', '/#'
sub Parse_MacroLines($\@);
sub Parse_MacroLines($\@) {
	my ($depth, $lines) = @_;
	my @list = ();
	while (defined(my $line = shift(@$lines))) {
		next if ($line =~ /^%/);
		my ($indent, $content) = $line =~ /^(\t*)([^\t].*)$/;
		$indent = length($indent);
		if ($indent < $depth) {
			unshift(@$lines, $line);
			last;
		}
		my ($title, $menukey, $text) = split("\t", $content, 3);
		if (defined($text)) { # menu item
			my %item = ( 'title' => $title, 'text' => DecodeSlashRT($text) );
			$item{'menukey'} = $menukey if (length($menukey));
			push(@list, \%item);
		} else { # hiermenu
			my $items = Parse_MacroLines($depth + 1, @$lines);
			my %hier = ( 'title' => $title );
			$hier{'list'} = $items if (@$items);
			push(@list, \%hier);
		}
	}
	return \@list;
}
sub Parse_MacroMenu() {
	my @files = Get_Files("MacroMenu", "MacroMenus Folder");
	return undef unless @files;
	my @list = ();
	for (@files) {
		my @lines = ReadFile($_);
		my $lst = Parse_MacroLines(0, @lines);
		push(@list, @$lst);
	}
	TrashFile("MacroMenus Folder") if (-e "$Root/MacroMenus Folder");
	return {
		'help' => '現在の選択部分を /# におき，/| を選択する．',
		'list' => \@list
	};
}

# MathRange
#	^.*$
sub Parse_MathRange() {
	return undef unless(-e "$Root/MathRange");
	my @lines = ReadFile("MathRange");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		if ($line =~ /^(.+)\t(.+)$/) {
			push(@list, { 'begin' => $1, 'end' => $2 });
		}
	}
	return { 'list' => \@list };
}

# MenuKey
# (.+)\t(.+)
sub Parse_MenuKey() {
	return undef unless(-e "$Root/MenuKey");
	my @lines = ReadFile("MenuKey");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		if ($line =~ /^(.+)\t(.+)$/) {
			push(@list, { 'title' => $1, 'menukey' => $2 });
		}
	}
	return {
		'help' => '存在しない項目は無視されるので英語版と兼用',
		'list' => \@list
	};
}

# ParallelFind
# (.+)\t(.+)
sub Parse_ParallelFind() {
	return undef unless(-e "$Root/ParallelFind");
	my @lines = ReadFile("ParallelFind");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		if ($line =~ /^(.+)\t(.+)$/) {
			push(@list, { 'find' => $1, 'replace' => $2 });
		}
	}
	return {
		'list' => \@list
	};
}

# PerlFind
sub Parse_PerlFind() {
	return undef unless(-e "$Root/PerlFind");
	my @lines = ReadFile("PerlFind");
	my $text = join("\n", @lines) . "\n";
	return {
		'text' => $text
	};
}

# SkipSpell
sub Parse_SkipSpell() {
	return undef unless(-e "$Root/SkipSpell");
	my @lines = ReadFile("SkipSpell");
	my $text = join("\n", @lines) . "\n";
	return {
		'text' => $text
	};
}

# TemplateMenu
# title\s+(.+)\t(.*)
# hiermenu\s+(.+)
# endhiermenu
sub Parse_TemplateText(\@) {
	my ($lines) = @_;
	my $text = '';
	my @next = ();
	while (defined(my $line = shift(@$lines))) {
		if ($line =~ /^title\s+([^\t]+)(?:\t(.+))?$/) {
			@next = ('title', $1, $2);
			last;
		} elsif ($line =~ /^hiermenu\s+(.+)$/) {
			@next = ('hiermenu', $1);
			last;
		} elsif ($line =~ /^endhiermenu(\s|$)/) {
			@next = ('endhiermenu');
			last;
		} else {
			$text .= "$line\n";
		}
	}
	my $len = length($text);
	if ($len >= 2 and substr($text, -2) eq "\n\n") {
		$text = substr($text, 0, $len - 1);
	}
	return ($text, @next);
}
sub Parse_TemplateLines(\@);
sub Parse_TemplateLines(\@) {
	my ($lines) = @_;
	my @list = ();
	# ignore head comments
	my ($text, $key, $title, $menukey) = Parse_TemplateText(@$lines);
	while ($key) {
		if ($key eq 'title') {
			my %hash = ( 'title' => $title );
			$hash{'menukey'} = $menukey if ($menukey);
			($text, $key, $title, $menukey) = Parse_TemplateText(@$lines);
			$hash{'text'} = $text;
			push(@list, \%hash);
		} elsif ($key eq 'hiermenu') {
			my %hash = ( 'title' => $title, 'list' => Parse_TemplateLines(@$lines) );
			push(@list, \%hash);
			($text, $key, $title, $menukey) = Parse_TemplateText(@$lines);
		} elsif ($key eq 'endhiermenu') {
			return \@list;
		}
	}
	return \@list;
}
sub Parse_TemplateMenu() {
	my @files = Get_Files("TemplateMenu", "TemplateMenus Folder");
	return undef unless @files;
	my @list = ();
	for (@files) {
		my @lines = ReadFile($_);
		my $lst = Parse_TemplateLines(@lines);
		push(@list, @$lst);
	}
	TrashFile("TemplateMenus Folder") if (-e "$Root/TemplateMenus Folder");
	return { 'list' => \@list };
}

# TeXAutoDelete
sub Parse_TeXAutoDelete() {
	return undef unless(-e "$Root/TeXAutoDelete");
	my @lines = ReadFile("TeXAutoDelete");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		push(@list, $line);
	}
	return { 'list' => \@list };
}

# TeXErrorMark
#	decode '/r', '/t'
sub Parse_TeXErrorMark() {
	return undef unless(-e "$Root/TeXErrorMark");
	my @lines = ReadFile("TeXErrorMark");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		my ($key, $lineno, $sign) =
		$line =~ /^([^\t]+)\t([^\t]+)(?:\t([-+]))?$/;
		if ($lineno) {
			$sign = ($sign eq '+') ? 'hilite' : 'menu';
			my %hash = (
				'key' => DecodeSlashRT($key),
				'line' => DecodeSlashRT($lineno),
				'type' => $sign
			);
			push(@list, \%hash);
		}
	}
	return { 'list' => \@list };
}

# TeXNoAutoOpen
sub Parse_TeXNoAutoOpen() {
	return undef unless(-e "$Root/TeXNoAutoOpen");
	my @lines = ReadFile("TeXNoAutoOpen");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		push(@list, $line);
	}
	return { 'list' => \@list };
}

# UnixTeXViewer
sub Parse_UnixTeXViewer() {
	return undef unless(-e "$Root/UnixTeXViewer");
	my @lines = ReadFile("UnixTeXViewer");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		push(@list, $line);
	}
	return {
		'help' => 'MacpTeX フォルダと OSX の Applications フォルダは自動検索します．
それ以外の場合は Helper フォルダにエイリアスを入れて下さい．
menu *** はメニュー項目で処理します．
shell *** は .dvi を MiniTerminal で処理します．',
		'list' => \@list
	};
}

# VarietyMenu
sub Parse_VarietyMenu() {
	return undef unless(-e "$Root/VarietyMenu");
	my @lines = ReadFile("VarietyMenu");
	my @list = ();
	for my $line (@lines) {
		next if ($line =~ /^%/);
		push(@list, $line);
	}
	return {
		'help' => '同じ項目名があるときは " : " で経路を指定すること．
存在しない項目は無視されるので英語版と兼用．',
		'list' => \@list
	};
}

# ScriptMenu
#	keywords:
#	endofscript,
#	hiermenu "", endhiermenu, hiermenu_j "", hiermenu_e ""
#	title "", title_j "", title_e ""
#	key ""
#	applescript $-$, perl [] $-$, shell [] $-$, minishell [] $-$
my %ScriptMenu_KeyWord = (
	'endofscript' => 'endofscript',
	'hiermenu' => 'hiermenu',
	'hiermenu_j' => 'hiermenu',
	'hiermenu_e' => 'hiermenu',
	'endhiermenu' => 'hiermenu',
	'title' => 'title',
	'title_j' => 'title',
	'title_e' => 'title',
	'key' => 'menukey',
	'applescript' => 'script',
	'perl' => 'script',
	'shell' => 'script',
	'minishell' => 'script'
);
sub ScriptMenu_ParseLine($) {
	my ($line) = @_;
	return () if ($line =~ /^[%#]/);
	my ($head) = $line =~ /^(\w+)/ or return ();
	my $type = $ScriptMenu_KeyWord{$head};
	return () unless $type;
	if ($type eq 'endofscript' or $type eq 'script') {
		return ($type, $head);
	} elsif ($type eq 'hiermenu' or $type eq 'title' or $type eq 'menukey') {
		my ($name) = $line =~ /^\w+\s+(.*)$/;
		$name = $1 if ($name and $name =~ /^"(.*)"$/);
		return ($type, $head, $name);
	} else {
		return ();
	}
}
sub ScriptMenu_CopyAndKill(\@\%$) {
	my ($list, $hash, $text) = @_;
	my %h = %$hash;
	return unless (%h);
	$h{'text'} = $text if (defined $text);
	undef %$hash;
	push(@$list, \%h);
}
sub ScriptMenu_ParseLines(\@);
sub ScriptMenu_ParseLines(\@) {
	my ($lines) = @_;
	my @list = ();
	my %item = ();
	my $text = '';
	while (defined(my $line = shift @$lines)) {
		my ($type, $head, $name) = ScriptMenu_ParseLine($line);
		unless ($type) {
			$text .= "$line\n";
			next;
		}
		if ($type eq 'endofscript') {
			last;
		} elsif ($type eq 'title' and $head eq 'title') {
			ScriptMenu_CopyAndKill(@list, %item, $text); $text = '';
			$item{$head} = $name;
		} elsif ($type eq 'title') {
			$item{$head} = $name;
		} elsif ($type eq 'menukey') {
			$item{'menukey'} = $name;
		} elsif ($type eq 'script') {
			$text .= "$line\n";
			my $mark = shift @$lines;
			$text .= "$mark\n";
			$text .= "$line\n" while (($line = shift @$lines) ne $mark);
			$text .= "$mark\n";
		} elsif ($type eq 'hiermenu') {
			ScriptMenu_CopyAndKill(@list, %item, $text); $text = '';
			if ($head eq 'hiermenu') {
				$item{'title'} = $name;
				while (defined($line = shift @$lines)) {
					($type, $head, $name) = ScriptMenu_ParseLine($line);
					next unless ($type);
					if ($type eq 'title') { unshift(@$lines, $line); last; }
					$head =~ /^hiermenu_([je])/ or die;
					$item{"title_$1"} = $name;
				}
				my @lst = ScriptMenu_ParseLines(@$lines);
				$item{'list'} = \@lst;
				ScriptMenu_CopyAndKill(@list, %item, undef);
			} elsif ($head eq 'endhiermenu') {
				last;
			} else {
				die;
			}
		}
	}
	ScriptMenu_CopyAndKill(@list, %item, $text);
	return @list;
}
sub ScriptMenu_ParseFile($) {
	my ($file) = @_;
	my @lines = ReadFile($file);
	return ScriptMenu_ParseLines(@lines);
}
sub ScriptMenu_ParseDir($);
sub ScriptMenu_ParseDir($) {
	my ($dir) = @_;
	my %hash;
	$dir =~ m|/(\d+ )?([^/]+)$|;
	$hash{'title'} = $2 if defined($2);
	if (-e "$Root/$dir/_name_") {
		my @lines = ReadFile("$dir/_name_");
		for (@lines) {
			$hash{'title' . ($1 ? $1 : '')} = $2 if (/^name(_[je])?\s(.*)$/)
		}
	}
	my @files = Get_Files('', $dir);
	my @list = ();
	for (@files) {
		next if (/^\./);
		next if ($_ eq '_name_');
		if (-d "$Root/$_") { push(@list, ScriptMenu_ParseDir($_)); }
		else { push(@list, ScriptMenu_ParseFile($_)); }
	}
	$hash{'list'} = \@list;
	return \%hash;
}
sub Parse_ScriptMenu() {
	my @files = Get_Files("ScriptMenu", "ScriptMenus Folder");
	my @list = ();
	for (@files) {
		if (-d "$Root/$_") { push(@list, ScriptMenu_ParseDir($_)); }
		else { push(@list, ScriptMenu_ParseFile($_)); }
	}
	TrashFile("ScriptMenus Folder") if (-e "$Root/ScriptMenus Folder");
	return { 'list' => \@list };
}

# main
{
	my %dict = ();
	if (-e "$Root/TextPref.plist") {
		my $plist = PListFileToPerl("$Root/TextPref.plist");
		%dict = %$plist;
	}
	AddToDict(%dict, 'IndexMenu', Parse_IndexMenu());
	AddToDict(%dict, 'CloseSymbol', Parse_CloseSymbol());
	AddToDict(%dict, 'ContextEditMenu', Parse_ContextEditMenu());
	AddToDict(%dict, 'MacroMenu', Parse_MacroMenu());
	AddToDict(%dict, 'MathRange', Parse_MathRange());
	AddToDict(%dict, 'MenuKey', Parse_MenuKey());
	AddToDict(%dict, 'ParallelFind', Parse_ParallelFind());
	AddToDict(%dict, 'PerlFind', Parse_PerlFind());
	AddToDict(%dict, 'SkipSpell', Parse_SkipSpell());
	AddToDict(%dict, 'TemplateMenu', Parse_TemplateMenu());
	AddToDict(%dict, 'TeXAutoDelete', Parse_TeXAutoDelete());
	AddToDict(%dict, 'TeXErrorMark', Parse_TeXErrorMark());
	AddToDict(%dict, 'TeXNoAutoOpen', Parse_TeXNoAutoOpen());
	AddToDict(%dict, 'UnixTeXViewer', Parse_UnixTeXViewer());
	AddToDict(%dict, 'VarietyMenu', Parse_VarietyMenu());
	AddToDict(%dict, 'ScriptMenu', Parse_ScriptMenu());
	PerlObjToPListFile("$Root/TextPref.plist", \%dict);
}
