package ExpandTeX;
use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(Expand SetNumArgs Define Undefined Eval);

# 2000/11/21 by N. KOISO

=usage
			
ExpandTeX::Expand($text)
	Expands $text and returns the expanded text.
	$text is a TeX format text.

ExpandTeX::SetNumArgs($num, @cs)
	Sets the number of arguments of macros.
	ExpandTeX::Expand does not expand these macros.
	$num is the number of arguments.
	@cs is a list of macros.
			
ExpandTeX::Define($text)
	Defines simple macros.
	$text is a TeX format text.
	Simple macros are as follows.
	("\def\cs#1,#2 {..}" is not <simple>.)
		\let\name\other
		\let\name=\other
		\def\name{other}
		\def\name#1#2#3{#1 and #2 and #3}
		\newcommand{\name}{other}
		\renewcommand{\name}{other}
		\newcommand{\name}[2]{#1 and #2}
		\renewcommand{\name}[2]{#1 and #2}
		
ExpandTeX::Undefined()
	Returns a list of unknown macros.
	These are assumed to have no arguments in ExpandTeX::Expand.
	When it is not empty, add them in SetNumArgs, and re-Expand.

ExpandTeX::Eval($text)
	Eval $text as a Perl script.
	Use it to define complex macros, or to suppress expansion of macros.
	See ExpandSample.xtra.
	Usuful subs:
		SetMacro($cs, &sub)
			$cs is expanded as &sub.
			sub should be a anonymous subroutine of form:
			sub($name, $in)
				$name is $cs. usually not used.
				$in is the reference to the TeX text follows to $cs.
				Should return the replacement.
				Sub is evaluated in ExpandTeX::.
				Use GetArgs($$in, $num) to receive is arguments.
		SetNumArgs($num, @cs)
			See above.

=cut

my %CSComplex = ();
my %CSReplace = ();
my %CSNumArgs = ();

sub GetToken(\$) {
	my ($in) = @_;
	return '' unless length($$in);
	$$in =~ s/^(\\([a-zA-Z]+|[^a-zA-Z])|\%.*\n?|[^\\\%])//;
	return $1;
}

sub GetParam(\$) {
	my ($in) = @_;
	return '' unless length($$in);
	my $out = GetToken($$in);
	return $out unless ($out eq '{');
	my $depth = 1;
	while (length($$in)) {
		$$in =~ s/^(([^\\\%\{\}]|\\.?|\%[^\n]*\n?)*)//s;
		$out .= $1;
		$$in =~ s/^(\{|\})//;
		$depth += ($1 eq '{') ? 1 : -1;
		$out .= $1;
		return $out if ($depth <= 0);
	}
	die "\nError (GetParam: unmatch {, }) at $out\n$!\n";
}

sub CSAlpha(\$) {
	my ($in) = @_;
	return '' unless (length($$in));
	return '' if (substr($$in, -1) !~ /[a-zA-Z]/);
	$$in =~ /(\\*)[a-zA-Z]*$/;
	return (length($1) % 2) == 1;
}

sub GetArgs(\$$) {
	my ($in, $num) = @_;
	my @out = ();
	my $last = '';
	while (--$num >= 0) {
		$$in =~ s/^\s*//;
		my $tok = $last = GetParam($$in);
		$tok =~ s/^\{(.*)\}$/$1/;
		push @out, $tok;
	}
	return (@out, $last);
}

sub CSAppend(\$\$) {
	my ($p, $q) = @_;
	my $m = ($$q =~ /^[a-zA-Z]/ and CSAlpha($$p)) ? ' ' : '';
	$$p .= $m.$$q;
}

sub ParseSimple($\$) {
	my ($name, $in) = @_;
	my $replace = $CSReplace{$name};
	my $last = $name;
	if ((my $num = $CSNumArgs{$name}) > 0) {
		my @params = GetArgs($$in, $num);
		unshift @params, '';
		for (my $i = 1; $i <= $num; ++$i) {
			my $r = $params[$i];
			while (my ($p, $q) = $replace =~ /^(.*?)\#$i(.*)$/) {
				CSAppend($p, $r);
				CSAppend($p, $q);
				$replace = $p;
			}
		}
		$last = $params[-1];
	}
	if ($last =~ /^\\.*[a-zA-Z]$/) {
		$$in =~ s/^ *// if ($$in =~ /^\s/ and ! CSAlpha($replace));
		# "\cs a" -> "\|a"; but '\t' and '\r' are preferred
	} elsif ($$in =~ /^[a-zA-Z]/ and CSAlpha($replace)) {
		$replace .= ' ' ; # "\|p" -> "\cs p"
	}
	return $replace;
}

sub Cover($) { # add '{', '}' if not single token
	my ($in) = @_;
	return "{$in}" if ($in =~ /^\s*$/);
	return $in if (length($in) <= 1);
	return "{$in}" if (substr($in, -1) eq "\n");
	return $in if ($in =~ /^\\([a-zA-Z]+|.)$/);
	return "{$in}" if ($in !~ /^\{/);
	my $copy = $in;
	my $out = GetParam($copy);
	return length($copy) ? "{$in}" : $in;
}

sub Ungroup($);
sub Ungroup($) {
	my ($in) = @_;
	my $out = '';
	for (;;) {
		$out .= $1 if ($in =~ s/^(([^\\\{\%]|\\.)+)//);
		last unless (length($in));
		if ($in =~ s/^(\%.*\n)//) {
			$out .= $1;
		} else {
			my $g = GetParam($in);
			$g =~ /^\{(.*)\}$/;
			$g = Cover(Ungroup($1));
			CSAppend($out, $g);
		}
	}
	return $out;
}

sub Expand($);

sub ParseParam(\$$) {
	my ($in, $nargs) = @_;
	my $out = '';
	while ($nargs-- > 0) {
		$out .= $1 if ($$in =~ s/^(\s+)//);
		my $p = GetParam($$in);
		my $q = Expand($p);
		$q = Cover($q) if ($q ne $p);
		CSAppend($out, $q);
	}
	return $out;
}

sub DebugObj($) {
	my ($obj) = @_;
	return (defined $obj) ? $obj : 'UNDEFINED';
}

sub Expand($) {
	my ($in) = @_;
	my $out = '';
	for (;;) {
		$out .= $1 if ($in =~ s/^([^\\\%^_]+)//);
		last unless length($in);
		if ($in =~ s/^(\%.*\n)//) {
			$out .= $1;
		} elsif ($in =~ s/^([_^])//) {
			$out .= $1;
			$out .= ParseParam($in, 1);
		} else {
			my $name = GetToken($in);
			die "\nError (Expand: cs) at $name$in.\n$!\n" if ($name !~ /^\\/);
			my $replace;
			if (defined($CSReplace{$name})) {
				$replace = ParseSimple($name, $in);
			} elsif (defined($CSComplex{$name})) {
				$replace = &{$CSComplex{$name}}($name, \$in);
			} else {
				$out .= $name;
				my $nargs = defined($CSNumArgs{$name})
					? $CSNumArgs{$name} : ($CSNumArgs{$name} = -1);
				if ($nargs > 0) {
					my $p = ParseParam($in, $nargs);
					CSAppend($out, $p);
				}
				next;
			}
			$replace = Ungroup($replace);
			CSAppend($replace, $in);
			$in = $replace;
			$out .= ' ' if ($in =~ /^[a-zA-Z]/ and CSAlpha($out));
		}
	}
	return $out;
}

sub SetNumArgs($@) {
	my ($num, @cs) = @_;
	while (my $cs = shift @cs) {
		$CSNumArgs{$cs} = $num;
		undef $CSReplace{$cs};
	}
}

sub Define($) {
	my ($in) = @_;
	while (length($in)) {
		next if ($in =~ s/^(\s+|\%.*\n\s*)//);
		my $tok = GetToken($in);
		my ($type) = $tok =~ s/^\\(def|let|newcommand|renewcommand)$//
		or die "\nERROR (Define: format) at $tok$in\n$!\n";
		$in =~ s/^\s*//;
		my $name = GetParam($in);
		$name =~ s/^\{(.*)\}$/$1/;
		die "\nERROR (Define: cs) at \\$type$name$in\n$!\n" if ($name !~ /^\\/);
		$in =~ s/^\s*//;
		$tok = 0;
		if ($in =~ s/^(#\d)*#(\d)//) { $tok = $2; }
		elsif ($in =~ s/^\[(\d)\]//) { $tok = $1; }
		$CSNumArgs{$name} = $tok;
		$in =~ s/^\s*//;
		if ($type eq 'let') {
			$in =~ s/^=//;
			$tok = GetToken($in);
		} else {
			$tok = GetParam($in);
			$tok =~ s/^\{(.*)\}$/$1/;
		}
		$CSReplace{$name} = $tok;
	}
}

sub Undefined() {
	return grep($CSNumArgs{$_} < 0, keys %CSNumArgs);
}

sub Eval($) {
	my ($in) = @_;
	eval $in;
}

sub SetMacro($&) {
	my ($cs, $sub) = @_;
	$CSComplex{$cs} = $sub;
	undef $CSReplace{$cs};
}

1;
